#include <algorithm>
#include <msp/io/print.h>
#include <msp/strings/utils.h>
#include "binarypackage.h"
#include "builder.h"
#include "filetarget.h"

using namespace std;
using namespace Msp;

BinaryPackage::BinaryPackage(Builder &b, const string &n):
	Package(b, n)
{
	use_pkgconfig = false;
}

void BinaryPackage::do_prepare()
{
	bool has_relative_paths = false;
	for(BuildInfo::PathList::const_iterator i=export_binfo.libpath.begin(); (!has_relative_paths && i!=export_binfo.libpath.end()); ++i)
		has_relative_paths = !i->is_absolute();
	for(BuildInfo::PathList::const_iterator i=export_binfo.incpath.begin(); (!has_relative_paths && i!=export_binfo.incpath.end()); ++i)
		has_relative_paths = !i->is_absolute();

	list<FS::Path> bases;

	/* If we have any relative paths that need resolving, or we have no paths at
	all and are not using pkg-config, look for files in prefix */
	if(has_relative_paths || (!use_pkgconfig && export_binfo.libpath.empty() && export_binfo.incpath.empty()))
		bases.push_back(builder.get_prefix());

	// Always look in system locations
	bases.push_back(FS::Path());

	bool system = false;
	for(list<FS::Path>::const_iterator i=bases.begin(); i!=bases.end(); ++i)
	{
		FS::Path prefix = *i;
		system = prefix.empty();
		if(system)
		{
			prefix = "/usr";
			const Architecture &arch = builder.get_current_arch();
			if(arch.is_cross())
				prefix /= arch.get_cross_prefix();
		}

		BuildInfo::PathList libpath = export_binfo.libpath;
		if(!system && libpath.empty())
			libpath.push_back("lib");
		for(BuildInfo::PathList::iterator j=libpath.begin(); j!=libpath.end(); ++j)
			*j = prefix/ *j;

		bool all_found = true;
		for(BuildInfo::WordList::const_iterator j=export_binfo.libs.begin(); j!=export_binfo.libs.end(); ++j)
			all_found &= (builder.get_vfs().find_library(*j, libpath, export_binfo.libmode, system)!=0);

		BuildInfo::PathList incpath = export_binfo.incpath;
		if(!system && incpath.empty())
			incpath.push_back("include");
		for(BuildInfo::PathList::iterator j=incpath.begin(); j!=incpath.end(); ++j)
			*j = prefix/ *j;

		for(HeaderList::const_iterator j=headers.begin(); j!=headers.end(); ++j)
			all_found &= (builder.get_vfs().find_header(*j, 0, incpath, system)!=0);

		if(all_found)
		{
			base_path = prefix;
			builder.get_logger().log("configure", format("%s found in %s", name, ((system && use_pkgconfig) ? "system" : base_path.str())));
			break;
		}
	}

	if(base_path.empty())
	{
		// TODO report which files were not found
		problems.push_back("Cannot locate files");
		return;
	}

	/* Add default entries to paths if they're empty and the package was found
	in a non-system location */
	if(!system && export_binfo.incpath.empty())
		export_binfo.incpath.push_back(base_path/"include");
	if(!system && export_binfo.libpath.empty())
		export_binfo.libpath.push_back(base_path/"lib");

	if(has_relative_paths)
	{
		for(BuildInfo::PathList::iterator i=export_binfo.incpath.begin(); i!=export_binfo.incpath.end(); ++i)
			*i = base_path/ *i;
		for(BuildInfo::PathList::iterator i=export_binfo.libpath.begin(); i!=export_binfo.libpath.end(); ++i)
			*i = base_path/ *i;
	}

	if(!static_binfo.libs.empty())
	{
		BuildInfo::PathList combined_libpath = static_binfo.libpath;
		combined_libpath.insert(combined_libpath.end(), export_binfo.libpath.begin(), export_binfo.libpath.end());

		Target::Dependencies static_dep_libs;
		for(BuildInfo::WordList::const_iterator i=static_binfo.libs.begin(); i!=static_binfo.libs.end(); ++i)
			if(Target *lib = builder.get_vfs().find_library(*i, combined_libpath, BuildInfo::STATIC, system))
				static_dep_libs.push_back(lib);

		for(BuildInfo::WordList::const_iterator i=export_binfo.libs.begin(); i!=export_binfo.libs.end(); ++i)
			if(Target *lib = builder.get_vfs().find_library(*i, export_binfo.libpath, BuildInfo::FORCE_STATIC, system))
			{
				for(Target::Dependencies::const_iterator j=static_dep_libs.begin(); j!=static_dep_libs.end(); ++j)
					lib->add_transitive_dependency(**j);
			}
	}
}

BinaryPackage *BinaryPackage::from_flags(Builder &builder, const string &name, const Flags &flags, const Flags &static_flags)
{
	BinaryPackage *pkg = new BinaryPackage(builder, name);
	pkg->use_pkgconfig = true;

	process_flags(flags, pkg->export_binfo);

	Flags exclusive_static_flags;
	for(Flags::const_iterator i=static_flags.begin(); i!=static_flags.end(); ++i)
		if(find(flags.begin(), flags.end(), *i)==flags.end())
			exclusive_static_flags.push_back(*i);
	process_flags(exclusive_static_flags, pkg->static_binfo);

	return pkg;
}

void BinaryPackage::process_flags(const Flags &flags, BuildInfo &binfo)
{
	for(Flags::const_iterator i=flags.begin(); i!=flags.end(); ++i)
	{
		if(!i->compare(0, 2, "-I"))
			binfo.incpath.push_back(i->substr(2));
		else if(!i->compare(0, 2, "-D"))
		{
			string::size_type equals = i->find('=');
			if(equals!=string::npos)
				binfo.defines[i->substr(2, equals-2)] = i->substr(equals+1);
			else
				binfo.defines[i->substr(2)] = string();
		}
		else if(!i->compare(0, 2, "-L"))
			binfo.libpath.push_back(i->substr(2));
		else if(!i->compare(0, 2, "-l"))
			binfo.libs.push_back(i->substr(2));
		else if(*i=="-pthread")
			binfo.threads = true;
	}
}


BinaryPackage::Loader::Loader(BinaryPackage &p):
	DataFile::DerivedObjectLoader<BinaryPackage, Package::Loader>(p)
{
	add("build_info", &Loader::build_info);
	add("header",     &Loader::header);
}

void BinaryPackage::Loader::build_info()
{
	load_sub(obj.export_binfo);
}

void BinaryPackage::Loader::header(const string &h)
{
	obj.headers.push_back(h);
}
