#include <msp/core/maputils.h>
#include <msp/fs/dir.h>
#include <msp/fs/utils.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "architecture.h"
#include "builder.h"
#include "component.h"
#include "externaltask.h"
#include "gnucompiler.h"
#include "objectfile.h"
#include "sourcefile.h"
#include "sourcepackage.h"

using namespace std;
using namespace Msp;

namespace {

const char *cpus[] =
{
	"athlonxp", "athlon-xp",
	"armv7a",   "armv7-a",
	0
};

}

GnuCompiler::GnuCompiler(Builder &b, const Architecture &a, const string &t, const FS::Path &sysroot):
	Tool(b, a, t)
{
	if(!sysroot.empty())
	{
		build_info.sysroot = sysroot;
		system_path.push_back(sysroot/"usr/include");
	}
	else if(architecture->is_native())
		system_path.push_back("/usr/include");
	else
		system_path.push_back("/usr/"+architecture->get_cross_prefix()+"/include");
}

Target *GnuCompiler::create_target(const list<Target *> &sources, const string &)
{
	if(sources.size()!=1)
		throw invalid_argument("GnuCCompiler::create_target");
	SourceFile &source = dynamic_cast<SourceFile &>(*sources.front());
	ObjectFile *obj = new ObjectFile(builder, *source.get_component(), source);
	obj->set_tool(*this);
	return obj;
}

string GnuCompiler::create_build_signature(const BuildInfo &binfo) const
{
	string result = FS::basename(executable->get_path());
	if(!architecture->get_cpu().empty())
	{
		result += ",m";
		result += architecture->get_cpu();
	}
	result += ',';
	if(binfo.debug)
		result += 'g';
	if(binfo.optimize)
	{
		result += 'O';
		result += (binfo.optimize>0 ? '0'+binfo.optimize : 's');
	}
	return result;
}

void GnuCompiler::do_prepare()
{
	executable = builder.get_vfs().find_binary(command);
	if(executable)
	{
		ExternalTask::Arguments argv;
		argv.push_back(executable->get_path().str());
		argv.push_back("-dumpversion");

		builder.get_logger().log("auxcommands", format("Running %s", join(argv.begin(), argv.end())));
		try
		{
			version = strip(ExternalTask::run_and_capture_output(argv));
			builder.get_logger().log("tools", format("%s version is %s", FS::basename(executable->get_path()), version));
		}
		catch(const runtime_error &)
		{ }
	}
}

Task *GnuCompiler::run(const Target &target) const
{
	const ObjectFile &object = dynamic_cast<const ObjectFile &>(target);

	ExternalTask::Arguments argv;
	argv.push_back(executable->get_path().str());
	argv.push_back("-c");

	BuildInfo binfo;
	target.collect_build_info(binfo);

	string tag_for_std = (tag=="OBJC" ? "CC" : tag);
	if(binfo.standards.count(tag_for_std))
		argv.push_back("-std="+get_item(binfo.standards, tag_for_std));
	if(tag=="OBJC" && binfo.standards.count(tag))
		argv.push_back("-fobjc-std="+get_item(binfo.standards, tag));

	if(binfo.warning_level>=1)
	{
		argv.push_back("-Wall");
		if(binfo.warning_level>=2)
		{
			argv.push_back("-Wextra");
			argv.push_back("-Wundef");
		}
		if(binfo.warning_level>=3)
		{
			argv.push_back("-pedantic");
			argv.push_back("-Wno-long-long");
			argv.push_back("-Wshadow");
			if(tag=="CC")
			{
				argv.push_back("-Wc++-compat");
				argv.push_back("-Wstrict-prototypes");
			}
		}
		if(binfo.warning_level>=4)
		{
			// Some truly paranoid warnings
			argv.push_back("-Wstrict-overflow=4");
			argv.push_back("-Wfloat-equal");
			argv.push_back("-Wconversion");
			argv.push_back("-Wwrite-strings");
			argv.push_back("-Winline");
		}
		if(binfo.fatal_warnings)
		{
			argv.push_back("-Werror");
			argv.push_back("-Wno-error=deprecated-declarations");
		}
	}

	const FS::Path &sysroot = binfo.sysroot;
	if(!sysroot.empty())
		argv.push_back("--sysroot="+sysroot.str());
	for(BuildInfo::PathList::const_iterator i=binfo.local_incpath.begin(); i!=binfo.local_incpath.end(); ++i)
	{
		argv.push_back("-iquote");
		argv.push_back(i->str());
	}
	for(BuildInfo::PathList::const_iterator i=binfo.incpath.begin(); i!=binfo.incpath.end(); ++i)
		argv.push_back("-I"+i->str());

	for(BuildInfo::DefineMap::const_iterator i=binfo.defines.begin(); i!=binfo.defines.end(); ++i)
	{
		if(i->second.empty())
			argv.push_back(format("-D%s", i->first));
		else
			argv.push_back(format("-D%s=%s", i->first, i->second));
	}

	if(binfo.debug)
		argv.push_back("-ggdb");
	if(binfo.optimize)
	{
		if(binfo.optimize<0)
			argv.push_back("-Os");
		else
			argv.push_back(format("-O%d", binfo.optimize));
	}
	if(binfo.threads && architecture->get_system()!="windows" && architecture->get_system()!="darwin")
		argv.push_back("-pthread");
	if(object.is_used_in_shared_library() && architecture->get_system()!="windows")
		argv.push_back("-fPIC");

	const Architecture &native_arch = builder.get_native_arch();
	if(architecture->is_native() && architecture->get_bits()!=native_arch.get_bits())
		argv.push_back(format("-m%d", architecture->get_bits()));

	string cpu = architecture->get_cpu();
	if(!cpu.empty())
	{
		for(unsigned i=0; cpus[i]; i+=2)
			if(cpu==cpus[i])
			{
				cpu = cpus[i+1];
				break;
			}
		argv.push_back("-march="+cpu);
	}

	if(!architecture->get_fpu().empty())
	{
		if(architecture->get_type()=="x86")
		{
			argv.push_back("-mfpmath="+architecture->get_fpu());
			if(architecture->get_fpu()=="sse")
				argv.push_back("-msse2");
		}
		else if(architecture->get_type()=="arm")
		{
			argv.push_back("-mfpu="+architecture->get_fpu());
			argv.push_back("-mfloat-abi=softfp");
		}
	}

	FS::Path obj_path = object.get_path();
	FS::Path src_path = object.get_source().get_path();
	FS::Path work_dir = object.get_component()->get_package().get_source_directory();

	argv.push_back("-o");
	argv.push_back(relative(obj_path, work_dir).str());
	argv.push_back(relative(src_path, work_dir).str());

	return new ExternalTask(argv, work_dir);
}
