#include <algorithm>
#include <cstdlib>
#include <msp/core/maputils.h>
#include <msp/fs/utils.h>
#include <msp/io/print.h>
#include <msp/strings/lexicalcast.h>
#include <msp/strings/utils.h>
#include "androidapplicationcomponent.h"
#include "binarycomponent.h"
#include "binarypackage.h"
#include "builder.h"
#include "datapackcomponent.h"
#include "file.h"
#include "installcomponent.h"
#include "pkgconfigfile.h"
#include "sourcearchivecomponent.h"
#include "sourcegenerator.h"
#include "sourcepackage.h"
#include "tool.h"

using namespace std;
using namespace Msp;

SourcePackage::SourcePackage(Builder &b, const string &n, const FS::Path &f):
	Package(b, n),
	source_dir(FS::dirname(f)),
	build_type(0),
	config(*this),
	cache(*this)
{
	config.load();

	build_file = builder.get_vfs().get_target(f);
	if(!build_file)
		build_file = new File(builder, *this, f);
	source_archive = new SourceArchiveComponent(*this);
	components.push_back(source_archive);
}

SourcePackage::~SourcePackage()
{
	for(ComponentList::iterator i=components.begin(); i!=components.end(); ++i)
		delete *i;
}

FS::Path SourcePackage::get_temp_directory() const
{
	string subdir = builder.get_current_arch().get_name();
	if(build_type)
	{
		subdir += '.';
		subdir += build_type->get_name();
	}

	const FS::Path &temp = builder.get_temp_directory();
	if(temp.is_absolute())
		return temp/name/subdir;
	else
		return source_dir/temp/subdir;
}

FS::Path SourcePackage::get_output_directory() const
{
	const Architecture &arch = builder.get_current_arch();
	if(arch.is_native())
		return source_dir;
	else
		return source_dir/arch.get_name();
}

const Component &SourcePackage::get_component(const string &n) const
{
	for(ComponentList::const_iterator i=components.begin(); i!=components.end(); ++i)
		if((*i)->get_name()==n)
			return **i;
	throw key_error(n);
}

bool SourcePackage::match_feature(const string &feat, const string *comp) const
{
	string value = config.get_option("with_"+feat).value;
	if(comp)
		return value==*comp;
	else
		return lexical_cast<bool>(value);
}

void SourcePackage::set_build_type(const BuildType &t)
{
	build_type = &t;
}

void SourcePackage::do_prepare()
{
	BuildInfo final_build_info;

	if(build_type)
		final_build_info.update_from(build_type->get_build_info());

	final_build_info.update_from(build_info);
	build_info = final_build_info;

	build_info.incpath.push_back((builder.get_prefix()/"include").str());
	build_info.libpath.push_back((builder.get_prefix()/"lib").str());

	for(FeatureList::iterator i=features.begin(); i!=features.end(); ++i)
	{
		string ident = "WITH_"+toupper(i->name);
		string value = config.get_option("with_"+i->name).value;

		if(i->choices.empty())
		{
			if(!lexical_cast<bool>(value))
				continue;
			value = "1";
		}

		build_info.defines[ident] = value;
		if(i->exported)
			export_binfo.defines[ident] = value;
	}

	for(ComponentList::iterator i=components.begin(); i!=components.end(); ++i)
	{
		(*i)->prepare();
		(*i)->create_build_info();

		(*i)->update_exported_build_info(export_binfo);
	}

	cache.load();

	for(ComponentList::iterator i=components.begin(); i!=components.end(); ++i)
		(*i)->create_targets();

	if(!export_binfo.libs.empty())
	{
		export_binfo.incpath.push_back((builder.get_prefix()/"include").str());
		export_binfo.libpath.push_back((builder.get_prefix()/"lib").str());

		PkgConfigFile *pc = new PkgConfigFile(builder, *this);
		builder.get_build_graph().get_target("install")->add_dependency(*builder.get_toolchain().get_tool("CP").create_target(*pc));
	}
}

void SourcePackage::save_caches()
{
	config.save();
	cache.save();
}


SourcePackage::Loader::Loader(SourcePackage &p):
	DataFile::DerivedObjectLoader<SourcePackage, Package::Loader>(p),
	FeatureConditional(p, p.name)
{
	init(0);
}

SourcePackage::Loader::Loader(SourcePackage &p, const Config::InputOptions &o):
	DataFile::DerivedObjectLoader<SourcePackage, Package::Loader>(p),
	FeatureConditional(p, p.name)
{
	init(&o);
}

void SourcePackage::Loader::init(const Config::InputOptions *o)
{
	options = o;
	add("android_application", &Loader::component<AndroidApplicationComponent>);
	add("build_info",  &Loader::build_info);
	add("datapack",    &Loader::component<DataPackComponent>);
	add("description", &SourcePackage::description);
	add("feature",     &Loader::feature);
	add("generate",    &Loader::generate);
	add("install",     &Loader::component<InstallComponent>);
	add("interface_version", &Loader::interface_version);
	add("library",     &Loader::component_arg<BinaryComponent, BinaryComponent::Type, BinaryComponent::LIBRARY>);
	add("module",      &Loader::component_arg<BinaryComponent, BinaryComponent::Type, BinaryComponent::MODULE>);
	add("program",     &Loader::component_arg<BinaryComponent, BinaryComponent::Type, BinaryComponent::PROGRAM>);
	add("source_archive", &Loader::source_archive);
	add("source_tarball", &Loader::source_archive);
	add("tarball",     &Loader::tarball);
	add("version",     &Loader::version);
}

void SourcePackage::Loader::finish()
{
	/* Make sure the source tarball is last in the list so targets from all
	other components wil be created first */
	ComponentList::iterator i = find(obj.components.begin(), obj.components.end(), obj.source_archive);
	if(i!=obj.components.end())
		obj.components.splice(obj.components.end(), obj.components, i);
}

void SourcePackage::Loader::feature(const string &n, const string &d)
{
	Feature feat(n);
	feat.description = d;
	load_sub(feat);
	obj.features.push_back(feat);

	const Config::Option &opt = obj.config.add_option(feat);
	if(options)
	{
		Config::InputOptions::const_iterator i = options->find(opt.name);
		if(i!=options->end())
			obj.config.set_option(opt.name, i->second);
	}
}

template<typename C>
void SourcePackage::Loader::component(const string &n)
{
	C *comp = new C(obj, n);
	load_sub(*comp);
	obj.components.push_back(comp);
}

template<typename C, typename A, A a>
void SourcePackage::Loader::component_arg(const string &n)
{
	C *comp = new C(obj, n, a);
	load_sub(*comp);
	obj.components.push_back(comp);
}

void SourcePackage::Loader::build_info()
{
	load_sub(obj.build_info);
}

void SourcePackage::Loader::generate(const string &tag)
{
	SourceGenerator *gen = new SourceGenerator(obj.builder, obj, tag);
	load_sub(*gen);
	obj.local_tools.add_tool(gen);
}

void SourcePackage::Loader::interface_version(const string &v)
{
	obj.interface_version = v;
	if(obj.version.empty())
		obj.version = v;
}

void SourcePackage::Loader::source_archive()
{
	load_sub(*obj.source_archive);
}

void SourcePackage::Loader::tarball(const string &)
{
	IO::print("%s: Deprecated tarball component ignored\n", get_source());
}

void SourcePackage::Loader::version(const string &v)
{
	obj.version = v;

	string::size_type i = 0;
	for(unsigned dots=0; i<obj.version.size(); ++i)
		if(obj.version[i]=='.' && ++dots>=2)
			break;
	obj.interface_version = obj.version.substr(0, i);
}
